local mod = ...





-- *************************************************************************************************************
-- Custom Component: Security -------------------------------------------------------------------------------------
-- Calculates each building part's security score and creates security for consumption
-- *************************************************************************************************************
local COMP_SECURITY = {
	TypeName = "COMP_SECURITY",
	ParentType = "COMPONENT",
	Properties = {
		{ Name = "CurrentDay", Type = "integer", Default = 0, Flags = { "SAVE_GAME" } },
		{ Name = "VillageCenterAdded", Type = "boolean", Default = false, Flags = { "SAVE_GAME" } },
		{ Name = "SpawnList", Type = "list<string>", Default = nil, Flags = { "SAVE_GAME" } },
    }
}


-- -- addGenerator -------------------------------------------------------------------------------------
-- function COMP_SECURITY:addGenerator(objPart)

--     objPart:getOrCreateComponent("COMP_SECURITY_GENERATOR")
--     local compResGen = objPart:getOrCreateComponent("COMP_RESOURCE_GENERATOR")
--     compResGen.GrowRate = 0
--     compResGen:setResourceGeneratorData("BUILDING_FUNCTION_SECURITY_GENERATOR")
--     compResGen:setMaxQuantity(0)

-- end


-- Init -------------------------------------------------------------------------------------
function COMP_SECURITY:init()

    mod:log("COMP_SECURITY ______________________________________________________________________________________")


    -- Process security on loading the game
    if GetPopulation(self) >= MinPopCrime() then
        self:trackVillagers()
        self:trackWorkplaces()
    end

    -- Turn off all overlays
    self:getLevel():getComponentManager("COMP_SECURITY_OVERLAY"):getAllComponent():forEach(function(compOverlay)
        compOverlay.Show = false
    end)

    -- Ensure no buildings are stuck in edit mode (this can prevent the overlay from working at all)
    self:getLevel():getComponentManager("COMP_BUILDING"):getAllComponent():forEach(function(compBuilding)
        compBuilding.EditMode = false
        mod:log("Building edit mode turned off.")
    end)

    -- Daily script
    local compMainGameLoop = self:getLevel():find("COMP_MAIN_GAME_LOOP")
    event.register(self, compMainGameLoop.ON_NEW_DAY, function()

        -- Record the current day
        self.CurrentDay = self.CurrentDay + 1

        -- Give points to vanilla assets
        self:attachPoints("WoodenKeepBaseCoreRootPart", 10, 2)
        self:attachPoints("RusticLordHouseTowerRootPart", 2, 2)
        self:attachPoints("RusticLordHouseTowerRootPart2", 2, 2)
        self:attachPoints("RusticLordHouseCoreRootPart", 0, 2)
        self:attachPoints("RootPart2", 0, 2) -- wooden tower from lord manor
        self:attachPoints("RusticLordHouseExtension1RootPart", 0, 2)
        self:attachPoints("RusticLordHouseTowerRootPart2", 4, 2)
        self:attachPoints("RusticLordHouseTowerRootPart2", 4, 2)

        -- Calculate generated points from security buildings
        self:calcPoints()

        -- Execute security tracking code
        self:trackVillagers()
        self:trackWorkplaces()

    end)

end



-- Calculate Security Points -------------------------------------------------------------------------------------
function COMP_SECURITY:calcPoints()

    self:getLevel():getComponentManager("COMP_SECURITY_POINTS"):getAllEnabledComponent():forEach(function(compPoints)

        -- Calculate worker points based on employment
        local compWork = compPoints:getOwner():getEnabledComponent("COMP_WORKPLACE")
        local workerPoints = 0
        local workerCount = 0
        if (compWork) then

            local listWorker = compWork.AssignedWorkerList
            if (listWorker) then
                for i = 1, #listWorker do
                    if (listWorker[i]) then
                        workerCount = workerCount + 1
                    end
                end

                workerPoints = compPoints.WorkerPoints * workerCount

            end
        end

        -- Set the current Points based on base + worker points
        compPoints.Points = compPoints.BasePoints + workerPoints

        -- mod:log("Points generated by " .. tostring(compPoints:getOwner().Name) .. " = " .. tostring(compPoints.Points))

    end)

end



-- Track villagers -------------------------------------------------------------------------------------
function COMP_SECURITY:trackVillagers()

    local outlawPercent = GetOutlawCount(self) / GetPopulation(self)

    self:getLevel():getComponentManager("COMP_VILLAGER"):getAllEnabledComponent():forEach(function(compVillager)

        -- Track unemployed and make outlaws of those unemployed too long
        local compSecVill = compVillager:getOwner():getOrCreateComponent("COMP_SECURITY_VILLAGER")
        local job = compVillager:getJobInstance()
        local jobName = ""

        if (job) then
            jobName = job.AssetJob.JobName
            compSecVill.DayUnemployed = 0 -- villager is not unemployed
        else
            if (compSecVill.DayUnemployed == 0) then
                compSecVill.DayUnemployed = self.CurrentDay -- track when the villager become unemployed
            end
        end

        -- Generate chance for outlaws from unemployed villagers
        if (GetPopulation(self) >= MinPopCrime()) and (outlawPercent <= .2) and (self.CurrentDay > GraceDays("Worker")) and (compSecVill.DayUnemployed > 0) and (compSecVill.DayUnemployed < (self.CurrentDay - GraceDays("Worker"))) then
            self:createOutlawFromVillager(compVillager)
        end

        -- Reset any outlaws that have had their job manually changed
        if (compSecVill.OutlawFlag) then
            if not (jobName == "JOB_SECURITY_OUTLAW_NAME") then
                compVillager:changeJob("JOB_SECURITY_OUTLAW")
                jobName = "JOB_SECURITY_OUTLAW_NAME"
            end

        -- Reset any prisoners that have had their job manually changed
        elseif (compSecVill.PrisonerFlag) then
            if not (jobName == "JOB_SECURITY_PRISONER_NAME") then
                compVillager:changeJob("JOB_SECURITY_PRISONER")
                jobName = "JOB_SECURITY_PRISONER_NAME"
            end

        end

        -- Set the Defender flag for soldiers and guards
        if (jobName == "SOLDIER") or (jobName == "JOB_SECURITY_GUARD_NAME") then
            compSecVill.DefenderFlag = true
        else
            compSecVill.DefenderFlag = false
        end

    end)

end



-- Track Workplaces -------------------------------------------------------------------------------------
function COMP_SECURITY:trackWorkplaces()

    local pop = GetPopulation(self)
    local outlawPercent = GetOutlawCount(self) / pop

    -- Score workplaces and create outlaws in unprotected workplaces
    self:getLevel():getComponentManager("COMP_WORKPLACE"):getAllEnabledComponent():forEach(function(compWork)

        local compWorkName = compWork:getOwner().Name
        local compSecPart = compWork:getOwner():getOrCreateComponent("COMP_SECURITY_BUILDING_PART")

        -- mod:log(compWorkName .. ", parent = " .. compWork:getOwner():getParent().Name)

        local posPart = compWork:getOwner():getGlobalPosition()
        local score = self:getScore(posPart)
        if not score then score = 0 end

        -- mod:log(compWorkName .. ", score = " .. score)

        -- provide exemptions for basic workplaces that tend to be far away from town
        if (compWorkName == "ForesterPart" or compWorkName == "WoodCutterCampPart" or compWorkName == "StonecutterPart" or compWorkName == "HunterPart") then
            score = 10
        end

        -- find out if the score changed, for the grace period
        if (score == compSecPart.Score) then -- score is the same now as it was before
        else
            compSecPart.DayScoreChanged = self.CurrentDay -- record today as the day the score changed
        end

        -- record the current score
        compSecPart.Score = score

        -- mod:log("Workplace security score: " .. tostring(compWork:getOwner().Name) .. " = " .. tostring(score) .. ", Day Score Changed: " .. compSecPart.DayScoreChanged)


        -- Generate Outlaws where security score is too low after a grace period
        if (pop >= MinPopCrime()) and (outlawPercent <= 0.2) and (compSecPart.DayScoreChanged < (self.CurrentDay - GraceDays("Workplace"))) and (compSecPart.Score <= 5) then
            self:createOutlawFromWorkplace(compWork, compSecPart)
        end

        -- Create theft points near unprotected workplaces; these delete themselves after a few days
        local chance = RandomFloat(0.05, 0.75) * (compSecPart.Score + 1.25) -- creates a distribution where score = 0 results in ~60% chance of success, score = 5 is ~8%, score = 9 is <1%
        if (pop >= MinPopCrime()) and (compSecPart.Score < 10 and compSecPart.DaySpawnedTheftPoint < (self.CurrentDay - 2) and (chance <= 0.6)) then -- the building is vulnerable to theft and it's been at least 3 days since theft points were spawned (they have expired by now)

            local listWorker = compWork.AssignedWorkerList
            if (#listWorker > 0) then -- only spawn theft objects if there are workers, an empty workplace likely has no resources to steal

                -- Spawn theft object
                local pos = self:findTheftPosition(compWork:getOwner():getGlobalPosition())

                local i = math.random(1, 3)
                local objTheft = self:getLevel():createObject("PREFAB_SECURITY_OUTLAWTHEFTPOINT" .. i, pos)
                objTheft:setRotationY(2 * math.pi * math.random())

                local scaleX = RandomFloat(0.85, 1.15)
                local scaleY = RandomFloat(0.85, 1.15)
                objTheft:scale( { scaleX, scaleY, scaleX } )

                compSecPart.DaySpawnedTheftPoint = self.CurrentDay
                
                -- Assign a relevant resource to the theft point; if collected, this resource will be paid to the outlaws
                local compTheft = objTheft:getOrCreateComponent("COMP_SECURITY_OUTLAWTHEFTPOINT")
                local compBuilding = compWork:getOwner():findFirstParentWithComponent("COMP_BUILDING")

                if (compBuilding) then
                    local nameBuilding = compBuilding.AssetBuilding.Name
                    -- mod:log(nameBuilding)
                    if (string.find(nameBuilding, "BAKERY")) then
                        compTheft.Resource = "BREAD"
                    elseif (string.find(nameBuilding, "BLACKSMITH")) then
                        compTheft.Resource = "TOOL"
                    elseif (string.find(nameBuilding, "BREWERY")) then
                        compTheft.Resource = "BEER"
                    elseif (string.find(nameBuilding, "CHEESE_MAKER")) then
                        compTheft.Resource = "CHEESE"
                    elseif (string.find(nameBuilding, "COAL_HUT")) then
                        compTheft.Resource = "COAL"
                    elseif (string.find(nameBuilding, "COOPERAGE")) then
                        compTheft.Resource = "BARREL"
                    elseif (string.find(nameBuilding, "DAIRY_FARM")) then
                        compTheft.Resource = "MILK"
                    elseif (string.find(nameBuilding, "FISHING_HUT")) then
                        compTheft.Resource = "FISH"
                    elseif (string.find(nameBuilding, "GATHERING_HUT")) then
                        compTheft.Resource = "BERRIES"
                    elseif (string.find(nameBuilding, "GLASS_SMELTER")) then
                        compTheft.Resource = "GLASS"
                    elseif (string.find(nameBuilding, "GOLD_SMELTER")) then
                        compTheft.Resource = "GOLD_BARS"
                    elseif (string.find(nameBuilding, "HOP_FARM")) then
                        compTheft.Resource = "HOP"
                    elseif (string.find(nameBuilding, "HUNTERS_HUT")) then
                        compTheft.Resource = "BOAR"
                    elseif (string.find(nameBuilding, "IRON_SMELTER")) then
                        compTheft.Resource = "IRON"
                    elseif (string.find(nameBuilding, "JEWELER_WORKSHOP")) then
                        compTheft.Resource = "JEWELRY"
                    elseif (string.find(nameBuilding, "SAWMILL")) then
                        compTheft.Resource = "PLANK"
                    elseif (string.find(nameBuilding, "SHEEP_FARM")) then
                        compTheft.Resource = "WOOL"
                    elseif (string.find(nameBuilding, "STONECUTTER_CAMP")) then
                        compTheft.Resource = "STONE"
                    elseif (string.find(nameBuilding, "STONEMASON_HUT")) then
                        compTheft.Resource = "POLISHED_STONE"
                    elseif (string.find(nameBuilding, "TAILOR_WORKSHOP")) then
                        compTheft.Resource = "CLOTHES"
                    elseif (string.find(nameBuilding, "VINEYARD")) then
                        compTheft.Resource = "GRAPES"
                    elseif (string.find(nameBuilding, "WEAPONSMITH")) then
                        compTheft.Resource = "SWORD"
                    elseif (string.find(nameBuilding, "WEAVER_HUT")) then
                        compTheft.Resource = "WOOLEN_CLOTH"
                    elseif (string.find(nameBuilding, "WHEAT_FARM")) then
                        compTheft.Resource = "WHEAT"
                    elseif (string.find(nameBuilding, "WINDMILL")) then
                        compTheft.Resource = "FLOUR"
                    elseif (string.find(nameBuilding, "LUMBER_CAMP")) then
                        compTheft.Resource = "WOOD"
                    else
                        compTheft.Resource = "GOLD_COINS"
                    end
                else
                    compTheft.Resource = "GOLD_COINS"
                end

            end

        end

    end)


    -- Initiate a new outlaw camp if warranted
    local countOutlaw = 0
    self:getLevel():getComponentManager("COMP_SECURITY_VILLAGER"):getAllEnabledComponent():forEach(function(compSecVill)
        if (compSecVill.OutlawFlag) then
            countOutlaw = countOutlaw + 1
        end
    end)

    local countOutlawCamp = 0
    self:getLevel():getComponentManager("COMP_SECURITY_OUTLAWCAMP"):getAllEnabledComponent():forEach(function(compOutlawCamp)
        countOutlawCamp = countOutlawCamp + 1
    end)

    local demandOutlawCamp = math.ceil(countOutlaw / 5) -- round up
    if (demandOutlawCamp > countOutlawCamp) then -- there is demand above the current number
        for i = 1, (demandOutlawCamp - countOutlawCamp) do
            self:launchPathfinder("OUTLAWCAMP")
        end
    end
    
end



-- Get score from nearby security points -------------------------------------------------------------------------------------
function COMP_SECURITY:getScore(pos)

    local score = 0
    local dist = 0

    local objPoints = nil
    local posPoints = nil
    local points = 0

    self:getLevel():getComponentManager("COMP_SECURITY_POINTS"):getAllComponent():forEach(function(compPoints)

        objPoints = compPoints:getOwner()
        posPoints = objPoints:getGlobalPosition()

        dist = math.sqrt( (pos[1] - posPoints.x)^2 + (pos[3] - posPoints.z)^2 )
        -- mod:log("Distance: " .. tostring(objChildName) .. ", " .. tostring(dist))

        points = compPoints.Points

        if (points <= 2) then
            if (dist <= 30) then score = score + points end
        elseif (points <= 5) then
            if (dist <= 50) then score = score + points end
        elseif (points <= 10) then
            if (dist <= 70) then score = score + points end
        end

    end)

    return score

end



-- Attach Points to Vanilla assets -------------------------------------------------------------------------------------
function COMP_SECURITY:attachPoints(partName, basePoints, workerPoints)

    local compOverlay = nil
    local compPoints = nil
    local compWork = nil
    local listWorker = nil
    local countWorker = 0
    local pointsWorker = 0

    -- -- mod:log("Looking for parts: " .. tostring(partName))
    -- local objPart = self:getLevel():find(partName)

    -- compPoints = objPart:getOrCreateComponent("COMP_SECURITY_POINTS")
    -- compPoints.BasePoints = basePoints
    -- compPoints.WorkerPoints = workerPoints

    -- if (compPoints.Points == 0) then

    --     compWork = objPart:getEnabledComponent("COMP_WORKPLACE")
    --     if not (compWork) then
    --         compWork = objPart:getParent():getEnabledComponent("COMP_WORKPLACE")
    --     end

    --     if (compWork) then
    --         listWorker = compWork.AssignedWorkerList
    --         if (listWorker) then
    --             countWorker = 0
    --             for i = 1, #listWorker do
    --                 if (listWorker[i]) then
    --                     countWorker = countWorker + 1
    --                 end
    --             end
    --             pointsWorker = compPoints.WorkerPoints * countWorker
    --         end

    --     end

    --     compPoints.Points = compPoints.BasePoints + pointsWorker

    --     mod:log("Security points added to part: " .. partName)

    -- end


    local listPart = {}
    self:getLevel():find(partName, listPart)

    -- mod:log("listPart = " .. tostring(listPart))

    if (next(listPart, nil)) then
        for i = 1, #listPart do

            -- compOverlay = listPart[i]:getOrCreateComponent("COMP_SECURITY_OVERLAY")
            compOverlay = listPart[i]:getOrCreateComponent("COMP_SECURITY_OVERLAY_INIT")
            compPoints = listPart[i]:getOrCreateComponent("COMP_SECURITY_POINTS")
            compPoints.BasePoints = basePoints
            compPoints.WorkerPoints = workerPoints

            if (compPoints.Points == 0) then

                -- try this part or its parent to get a workplace (e.g. in the Wooden keep it is the parent part that has the workplace, not the part itself, but we want the points on the child part)
                compWork = listPart[i]:getEnabledComponent("COMP_WORKPLACE")
                if not (compWork) then
                    compWork = listPart[i]:getParent():getEnabledComponent("COMP_WORKPLACE")
                end

                if (compWork) then
                    listWorker = compWork.AssignedWorkerList
                    if (listWorker) then
                        countWorker = 0
                        for w = 1, #listWorker do
                            if (listWorker[w]) then
                                countWorker = countWorker + 1
                            end
                        end
                        pointsWorker = compPoints.WorkerPoints * countWorker
                    end

                end

                compPoints.Points = compPoints.BasePoints + pointsWorker

                -- mod:log("Security points added to part: " .. partName)

            end

        end
    end

end



-- Find theft position -------------------------------------------------------------------------------------
-- Looks for an empty (pathable) space near the workplace to attract outlaws
function COMP_SECURITY:findTheftPosition(posCenter)

    -- mod:log("Workplace centred at " .. tostring(posCenter))
    local posGuess = {0, 0, 0}

    -- try 100 times if needed
    local found = false
    for i = 1, 100 do

        posGuess[1] = math.random(posCenter[1] - 10, posCenter[1] + 10)
        posGuess[3] = math.random(posCenter[3] - 10, posCenter[3] + 10)

        -- mod:log("Guessing position at " .. tostring(posGuess))

        if not (HasBuildingCollision(self, posGuess)) and (GetWaterElevation(self, posGuess) < GetGroundElevation(self, posGuess)) then
            -- mod:log("Found a position that didn't have a collision or under water.")
            found = true
            break
        end

    end

    if (found) then
        return posGuess
    else
        return posCenter
    end

end



-- Launch pathfinder -------------------------------------------------------------------------------------
function COMP_SECURITY:launchPathfinder(type)

    local posVillageCenter = {500, 0, 500}

    -- find the village center
    self:getLevel():getComponentManager("COMP_BUILDING_PART"):getAllEnabledComponent():forEach(function(compPart)
        if compPart:getOwner().Name == "TempVillageCenterPart" then
            posVillageCenter = compPart:getOwner():getGlobalPosition()
            -- mod:log("Village center position found: " .. tostring(posVillageCenter))
        end
    end)

    local objPathfinder = self:getLevel():createObject("PREFAB_SECURITY_PATHFINDER", posVillageCenter)
    local compPathfinder = objPathfinder:getOrCreateComponent("COMP_SECURITY_PATHFINDER")
    if (type) then
        compPathfinder.Type = type
    end
    mod:log("Pathfinder agent launched at coordinates " .. tostring(posVillageCenter))

end



-- Create Outlaw from Workplace -------------------------------------------------------------------------------------
function COMP_SECURITY:createOutlawFromWorkplace(compWork, compSecPart)

    local rand = math.random() -- random float between zero and one
    local chance = rand * (compSecPart.Score + 1)

    local job = nil
    local nameJob = nil
    local banned = false

    -- mod:log("Undefended building part: " .. tostring(compPart:getOwner().Name) .. ", chance: " .. tostring(chance))

    if (chance < 0.1) then -- any workers will be outlaws if the score falls below n%

        -- get the workers
        local listWorker = compWork.AssignedWorkerList

        -- there is at least one worker
        if (listWorker[1]) then
            for i = 1, #listWorker do
                if (listWorker[i]) then

                    if not (listWorker[i]:getOwner():getOrCreateComponent("COMP_SECURITY_VILLAGER").OutlawFlag) then -- they haven't already been marked as an Outlaw from previous code

                        job = listWorker[i]:getAssetJob()
                        if (job) then
                            nameJob = job.JobName
                            if
                                nameJob == "SOLDIER"
                                or nameJob == "BAILIFF"
                                or nameJob == "JOB_SECURITY_CONSTABLE_NAME"
                                or nameJob == "JOB_SECURITY_GUARD_NAME"
                                or nameJob == "JOB_AUTHORITY_LORD_NAME"
                                or nameJob == "JOB_AUTHORITY_LADY_NAME"
                                then banned = true
                            end
                        end
                                            
                        if not banned then

                            -- trigger outlaw notification
                            listWorker[i]:changeJob("JOB_SECURITY_EVENT_OUTLAW")

                            -- change the job to Outlaw
                            listWorker[i]:changeJob("JOB_SECURITY_OUTLAW")

                            -- register a flag change the job back to Outlaw if it is changed by the player
                            listWorker[i]:getOwner():getOrCreateComponent("COMP_SECURITY_VILLAGER").OutlawFlag = true

                            mod:log("Villager " .. tostring(listWorker[i]:getOwner().Name) .. " became an Outlaw after working at an undefended workplace.")
                        
                        end

                    end
                end
            end
        end

    end

end



-- Create Outlaw from Unemployed Villager -------------------------------------------------------------------------------------
function COMP_SECURITY:createOutlawFromVillager(compVillager)

    local chance = math.random()

    if (chance < 0.25) then -- the villager will be an outlaw if their random number falls below n%

        -- trigger outlaw notification
        compVillager:changeJob("JOB_SECURITY_EVENT_OUTLAW")

        -- change the job to Outlaw
        compVillager:changeJob("JOB_SECURITY_OUTLAW")

        -- register a flag change the job back to Outlaw if it is changed by the player
        compVillager:getOwner():getOrCreateComponent("COMP_SECURITY_VILLAGER").OutlawFlag = true

        mod:log("Villager " .. compVillager:getOwner().Name .. " became an Outlaw after being unemployed for more than n days.")
    
    end

end



-- -- Spawn Object -------------------------------------------------------------------------------------
-- function COMP_SECURITY:spawnObject(prefab, pos, offset)

--     local objNew = self:getLevel():createObject(prefab, pos)
--     if (offset) then   
--         objNew:translate({offset, 0, 0})
--     end

--     local comp = objNew:getOrCreateComponent("COMP_RESOURCE_GENERATOR")
--     comp:setResourceGeneratorData("BUILDING_FUNCTION_SECURITY_RESOURCEGENERATOR")
--     comp:setMaxQuantity(1)
--     comp.GrowRate = 0

--     table.insert(self.SpawnList, objNew:getId())

-- end



-- destroySpawned -------------------------------------------------------------------------------------
function COMP_SECURITY:destroySpawned()

    -- remove any object in the spawn list 
    for i, idPair in ipairs(self.SpawnList) do
        local obj = self:getLevel():find(self.SpawnList[i])
        if (obj) then
            obj:destroyAllChild()
            obj:destroy()
        end
    end
    
    -- Clear the spawn list for future use
    for k,v in pairs(self.SpawnList) do
        self.SpawnList[k] = nil
    end

end



-- onDestroy -------------------------------------------------------------------------------------
function COMP_SECURITY:onDestroy(isClearingLevel)

    self:destroySpawned()
    
end



-- Register custom Class
mod:registerClass(COMP_SECURITY)

-- Attach class to prefab
mod:registerPrefabComponent("PREFAB_MANAGER", { DataType = "COMP_SECURITY" } )



